from typing import TypedDict
from .xml_parser import (
    XMLNode,
    XMLDoc,
)
from godocs.translation.translator import SyntaxTranslator
from godocs.translation.interpreter import Interpreter


class Property(TypedDict):
    name: str
    type: str
    default: str
    description: str


class Constant(TypedDict):
    name: str
    value: str
    description: str


class Method(TypedDict):
    name: str
    type: str
    args: list[Property]
    description: str


class Signal(TypedDict):
    name: str
    args: list[Property]
    description: str


class Enum(TypedDict):
    name: str
    values: list[Constant]
    description: str


class ThemeItem(TypedDict):
    name: str
    data_type: str
    type: str
    default: str
    description: str


class Class(TypedDict):
    name: str
    inheritage: list[str]
    brief_description: str
    description: str
    properties: list[Property]
    methods: list[Method]
    signals: list[Signal]
    constants: list[Constant]
    enums: list[Enum]
    theme_items: list[ThemeItem]


class DocContext(TypedDict):
    classes: list[Class]
    options: dict[str, str]


def get_class_node(class_name: str, docs: list[XMLDoc]) -> XMLNode | None:
    for doc in docs:
        root = doc.getroot()

        if root.attrib.get("name") == class_name:
            return root

    return None


def parse_inheritage(root: XMLNode, docs: list[XMLDoc]) -> list[str]:
    result: list[str] = []

    parent_name = root.attrib.get("inherits", '')

    while parent_name != '':
        result.append(parent_name)

        parent = get_class_node(parent_name, docs)

        if parent is None:
            break

        parent_name = parent.attrib.get("inherits", '')

    return result


def parse_property(node: XMLNode) -> Property:
    """
    Parses a member node into a dict.

    Member node structure::

        <member name="color" type="Color" setter="set_color" getter="get_color" default="null">
          Description
        </member>

    Return structure::

        result = {
            "name": "color",
            "type": "Color",
            "default": "null",
            "description": "Description",
        }
    """

    result: Property = {
        "name": '',
        "type": '',
        "default": '',
        "description": '',
    }

    result["name"] = node.attrib.get("name", '')
    result["type"] = node.attrib.get("type", '')
    result["default"] = node.attrib.get("default", '')
    result["description"] = node.text.strip() if node.text is not None else ''

    return result


def parse_method(node: XMLNode) -> Method:
    """
    Parses a method node into a dict.

    Method node structure::

        <method name="add">
          <return type="int" />
          <param index="0" name="num1" type="int" />
          <param index="1" name="num2" type="int" />
          <description>
            Description
          </description>
        </method>

    Return structure::

        result = {
          "name": "add",
          "type": "int",
          "args": [
            {
              "name": "num1",
              "type": "int",
              "default": "",
              "description": "",
            },
            {
              "name": "num2",
              "type": "int",
              "default": "",
              "description": "",
            },
          ],
          "description": "Description",
        }
    """

    result: Method = {
        "name": '',
        "type": '',
        "args": [],
        "description": '',
    }

    result["name"] = node.attrib.get("name", '')
    result["type"] = node.find("return").attrib.get("type", '')  # type: ignore
    result["description"] = node \
        .find("description") \
        .text.strip() if node.text is not None else ''  # type: ignore

    for arg in node.findall("param"):
        result["args"].append(parse_property(arg))

    return result


def parse_signal(node: XMLNode) -> Signal:
    """
    Parses a signal node into a dict.

    Signal node structure::

        <signal name="damaged">
          <param index="0" name="amount" type="float" />
          <description>
            Emitted when someone gets damaged.
          </description>
        </signal>

    Return structure::

        result = {
          "name": "damaged",
          "args": [
            {
              "name": "amount",
              "type": "float",
              "default": "",
              "description": "",
            },
          ],
          "description": "Emitted when someone gets damaged.",
        }
    """

    result: Signal = {
        "name": '',
        "args": [],
        "description": '',
    }

    result["name"] = node.attrib.get("name", '')
    result["description"] = node \
        .find("description") \
        .text.strip() if node.text is not None else ''  # type: ignore

    for arg in node.findall("param"):
        result["args"].append(parse_property(arg))

    return result


def parse_constant(node: XMLNode) -> Constant:
    """
    Parses a constant node into a dict.

    Constant node structure::

        <constant name="PI" value="3.14">
          The value of PI.
        </constant>

    Return structure::

        result = {
          "name": "PI",
          "value": 3.14,
          "description": "The value of PI.",
        }
    """

    result: Constant = {
        "name": '',
        "value": '',
        "description": '',
    }

    result["name"] = node.attrib.get("name", '')
    result["value"] = node.attrib.get("value", '')
    result["description"] = node.text.strip() if node.text is not None else ''

    return result


def parse_enum(name: str, values: list[XMLNode]) -> Enum:
    """
    Parses a list of constant nodes into an enum dict.

    Different from the other parse functions, this one accepts a
    name and a list of XMLNodes with the enum constants.

    This is done this way because the Godot XML docs don't
    treat enums as a separate "entity". Enum members are got from
    constants with a enum attribute in common, which defines the
    name from the enum.

    Also because of the XML docs structure, Enums don't store
    description data, which is why this function's enum dict
    will always have an empty str as its description.

    node structure::

        <constant name="ADDITION" value="0" enum="Operation">
          Operation of adding two numbers.
        </constant>
        <constant name="SUBTRACTION" value="1" enum="Operation">
          Operation of subtracting two numbers.
        </constant>

    Return structure::

        result = {
          "name": "Operation",
          "values": [
            {
              "name": "ADDITION",
              "value": "0",
              "description": "Operation of adding two numbers.",
            },
            {
              "name": "SUBTRACTION",
              "value": "1",
              "description": "Operation of subtracting two numbers.",
            },
          ],
          "description": "",
        }
    """

    result: Enum = {
        "name": '',
        "values": [],
        "description": '',
    }

    result["name"] = name

    for value in values:
        result["values"].append(parse_constant(value))

    # result["description"] =  There's no way to get enum descriptions with doctool

    return result


def parse_theme_item(node: XMLNode) -> ThemeItem:
    result: ThemeItem = {
        "name": '',
        "data_type": '',
        "type": '',
        "default": '',
        "description": '',
    }

    result["name"] = node.attrib.get("name", '')
    result["data_type"] = node.attrib.get("data_type", '')
    result["type"] = node.attrib.get("type", '')
    result["default"] = node.attrib.get("default", '')
    result["description"] = node.text.strip() if node.text is not None else ''

    return result


def parse_properties(node: XMLNode) -> list[Property]:
    """
    Parses a node with a list of properties into a dict.

    Properties node structure::

        <members>
          <member name="color" type="Color" setter="set_color" getter="get_color" default="null">
            A thinga with color and name.
          </member>
          <member name="name" type="String" setter="set_name" getter="get_name" default="null">
            The name of the thinga.
          </member>
        </members>

    Return structure::

        result = [
          {
            "name": "color",
            "type": "Color",
            "default": "null",
            "description": "A thinga with color and name.",
          },
          {
            "name": "name",
            "type": "String",
            "default": "null",
            "description": "The name of the thinga.",
          },
        ]
    """

    result: list[Property] = []

    for property in node:
        if property.text is None or property.text.strip() == '':
            continue

        result.append(parse_property(property))

    return result


def parse_methods(node: XMLNode) -> list[Method]:
    """
    Parses a node with a list of methods into a list of dicts.

    Methods node structure::

        <methods>
          <method name="add">
            <return type="int" />
            <param index="0" name="num1" type="int" />
            <param index="1" name="num2" type="int" />
            <description>
              Adds two numbers.
            </description>
          </method>
          <method name="subtract">
            <return type="int" />
            <param index="0" name="num1" type="int" />
            <param index="1" name="num2" type="int" />
            <description>
              Subtracts the second number from the first.
            </description>
          </method>
        </methods>

    Return structure::

        result = [
          {
            "name": "add",
            "type": "int",
            "args": [
              {"name": "num1", "type": "int", "default": "", "description": ""},
              {"name": "num2", "type": "int", "default": "", "description": ""},
            ],
            "description": "Adds two numbers.",
          },
          {
            "name": "subtract",
            "type": "int",
            "args": [
              {"name": "num1", "type": "int", "default": "", "description": ""},
              {"name": "num2", "type": "int", "default": "", "description": ""},
            ],
            "description": "Subtracts the second number from the first.",
          },
        ]
    """

    result: list[Method] = []

    for method in node.findall("method"):
        description = method.find("description").text  # type: ignore

        if description is None or description.strip() == '':
            continue

        result.append(parse_method(method))

    return result


def parse_signals(node: XMLNode) -> list[Signal]:
    """
    Parses a node with a list of signals into a list of dicts.

    Signals node structure::

        <signals>
          <signal name="damaged">
            <param index="0" name="amount" type="float" />
            <description>
              Emitted when someone gets damaged.
            </description>
          </signal>
          <signal name="healed">
            <param index="0" name="amount" type="float" />
            <description>
              Emitted when someone gets healed.
            </description>
          </signal>
        </signals>

    Return structure::

        result = [
          {
            "name": "damaged",
            "args": [
              {"name": "amount", "type": "float", "default": "", "description": ""},
            ],
            "description": "Emitted when someone gets damaged.",
          },
          {
            "name": "healed",
            "args": [
              {"name": "amount", "type": "float", "default": "", "description": ""},
            ],
            "description": "Emitted when someone gets healed.",
          },
        ]
    """

    result: list[Signal] = []

    for signal in node.findall("signal"):
        description = signal.find("description").text  # type: ignore

        if description is None or description.strip() == '':
            continue

        result.append(parse_signal(signal))

    return result


def parse_constants(node: XMLNode) -> list[Constant]:
    """
    Parses a node with a list of constants into a list of dicts.

    This function ignores the constants that have an enum attribute,
    which would make them members of an Enum.

    For getting members of an Enum, the parse_enums function should be used.

    Constants node structure::

        <constants>
          <constant name="PI" value="3.14">
            The value of PI.
          </constant>
          <constant name="E" value="2.71">
            The value of Euler's number.
          </constant>
        </constants>

    Return structure::

        result = [
          {
            "name": "PI",
            "value": "3.14",
            "description": "The value of PI.",
          },
          {
            "name": "E",
            "value": "2.71",
            "description": "The value of Euler's number.",
          },
        ]
    """

    result: list[Constant] = []

    for constant in node.findall("constant"):
        if constant.text is None or constant.text.strip() == '':
            continue
        if constant.attrib.get("enum", '') != '':
            continue

        result.append(parse_constant(constant))

    return result


def parse_enums(node: XMLNode) -> list[Enum]:
    """
    Parses a node with a list of constants into a list of enums.

    This function ignores the constants that don't have an enum attribute,
    which would make them just constants without an enum.

    For getting the solely constants, the parse_constants function should be used.

    Constants node structure::

        <constants>
          <constant name="ADDITION" value="0" enum="Operation">
            The addition operation.
          </constant>
          <constant name="SUBTRACTION" value="1" enum="Operation">
            The subtraction operation.
          </constant>
        </constants>

    Return structure::

        result = [
          {
            "name": "Operation",
            "values": [
              {
                "name": "ADDITION",
                "value": "0",
                "description": "The addition operation.",
              },
              {
                "name": "SUBTRACTION",
                "value": "1",
                "description": "The subtraction operation.",
              },
            ],
            "description": "",
          }
        ]
    """

    result: list[Enum] = []

    enums: dict[str, list[XMLNode]] = {}

    for constant in node.findall("constant"):
        enum_name = constant.attrib.get("enum", '')

        if enum_name == '':
            continue
        if constant.text is None or constant.text.strip() == '':
            continue

        if enums.get(enum_name) is None:
            enums[enum_name] = []

        enums[enum_name].append(constant)

    for enum in enums:
        values = enums[enum]

        result.append(parse_enum(enum, values))

    return result


def parse_theme_items(node: XMLNode) -> list[ThemeItem]:
    """
    Parses a node with a list of theme items into a list of dicts.

    Theme items node structure::

        <theme_items>
          <theme_item name="font_color" data_type="color" type="Color" default="black">
            The default font color.
          </theme_item>
          <theme_item name="font_size" data_type="int" type="int" default="14">
            The default font size.
          </theme_item>
        </theme_items>

    Return structure::

        result = [
          {
            "name": "font_color",
            "data_type": "color",
            "type": "Color",
            "default": "black",
            "description": "The default font color.",
          },
          {
            "name": "font_size",
            "data_type": "int",
            "type": "int",
            "default": "14",
            "description": "The default font size.",
          },
        ]
    """

    result: list[ThemeItem] = []

    for theme_item in node:
        if theme_item.text is None or theme_item.text.strip() == '':
            continue

        result.append(parse_theme_item(theme_item))

    return result


def parse_class(root: XMLNode, docs: list[XMLDoc]) -> Class:
    """
    Parses an XML node representing a Godot class into a convenient dict
    with separated information about the class members.

    The structure of the XML expected is the one generated by Godot's doctool
    and the generated dict has its structure defined in the Class type.
    """

    result: Class = {
        "name": '',
        "inheritage": [],
        "brief_description": '',
        "description": '',
        "properties": [],
        "methods": [],
        "signals": [],
        "constants": [],
        "enums": [],
        "theme_items": [],
    }

    result["name"] = root.attrib.get("name", '')
    result["inheritage"] = parse_inheritage(root, docs)
    result["brief_description"] = root \
        .find("brief_description").text.strip()  # type: ignore
    result["description"] = root \
        .find("description").text.strip()  # type: ignore

    for node in root:
        match node.tag:
            case "members": result["properties"] = parse_properties(node)
            case "methods": result["methods"] = parse_methods(node)
            case "signals": result["signals"] = parse_signals(node)
            case "constants":
                result["constants"] = parse_constants(node)
                result["enums"] = parse_enums(node)
            case "theme_items": result["theme_items"] = parse_theme_items(node)
            case _: pass

    return result


def create(
        docs: list[XMLDoc],
        options: dict[str, str] | None = None
) -> DocContext:
    """
    Creates a DocContext with information about all classes present in the
    docs list passed, as well as with data about the options desired to keep
    in context inside the options parameter.
    """

    if options is None:
        options = {}

    result: DocContext = {
        "options": options,
        "classes": [],
    }

    for doc in docs:
        root = doc.getroot()

        result["classes"].append(parse_class(root, docs))

    return result


def translate(ctx: DocContext, interpreter: Interpreter, translator: SyntaxTranslator):
    classes = ctx["classes"]
    for class_doc in classes:
        for member in class_doc:
            match(member):
                case "brief_description":
                    class_doc[member] = interpreter.interpret(
                        class_doc[member]).translate(translator)
                case "description":
                    class_doc[member] = interpreter.interpret(
                        class_doc[member]).translate(translator)
                case "constants":
                    constants = class_doc[member]
                    for constant in constants:
                        constant["description"] = interpreter.interpret(
                            constant["description"]).translate(translator)
                case "enums":
                    enums = class_doc[member]
                    for enum in enums:
                        enum["description"] = interpreter.interpret(
                            enum["description"]).translate(translator)
                        constants = enum["values"]
                        for constant in constants:
                            constant["description"] = interpreter.interpret(
                                constant["description"]).translate(translator)
                case "methods":
                    methods = class_doc[member]
                    for method in methods:
                        method["description"] = interpreter.interpret(
                            method["description"]).translate(translator)
                case "properties":
                    properties = class_doc[member]
                    for property in properties:
                        property["description"] = interpreter.interpret(
                            property["description"]).translate(translator)
                case "signals":
                    signals = class_doc[member]
                    for signal in signals:
                        signal["description"] = interpreter.interpret(
                            signal["description"]).translate(translator)
                case "theme_items":
                    theme_items = class_doc[member]
                    for theme_item in theme_items:
                        theme_item["description"] = interpreter.interpret(
                            theme_item["description"]).translate(translator)
                case _: pass

    return ctx



import json
from os import PathLike


def load(path: str | PathLike[str]):
    """
    Loads a **JSON file** located in `path` and
    parses its values to a `dict` variable.

    The loaded `dict` is then returned.
    """

    with open(path, "r", encoding="utf-8") as f:
        options = json.load(f)

    return options



from argparse import ArgumentParser, Namespace
from typing import Optional, TYPE_CHECKING
from godocs.cli.command.cli_command import CLICommand
from godocs.parser import xml_parser, context_creator
from godocs.translation.interpreter import BBCodeInterpreter
from godocs.translation.translator import get_translator
from godocs import util

if TYPE_CHECKING:
    from argparse import _SubParsersAction  # type: ignore
    from godocs.cli.command.cli_command import Processor


class ConstructCommand(CLICommand):
    """
    A `CLICommand` that allows choosing a `constructor` for
    generating documentation.
    """

    TRANSLATORS = ["rst"]
    """
    The default translators accepted by this command.
    """

    parser: ArgumentParser
    """
    The `argparse.ArgumentParser` instance this `ConstructCommand` uses.
    """

    subparsers: "_SubParsersAction[ArgumentParser]"
    """
    The `subparsers` of the `parser` of this `ConstructCommand`.
    """

    parent_parser: ArgumentParser = ArgumentParser(add_help=False)
    """
    An `ArgumentParser` that holds common parameters and options for all
    the constructor types chosen.
    """

    subcommands: dict[str, CLICommand] = {}
    """
    The subcommands this `ConstructCommand` exposes.
    """

    def register(
        self,
        superparsers: "Optional[_SubParsersAction[ArgumentParser]]" = None,
        parent_parser: Optional[ArgumentParser] = None,
        processors: "Optional[list[Processor]]" = None
    ):
        """
        Registers this `ConstructCommand` as a subparser for the
        `subparsers` received.
        """

        if superparsers is None:
            raise ValueError(
                'superparsers are needed for "construct" registration')
        if processors is None:
            raise ValueError(
                'processors are needed for "construct" registration')

        self.parser = superparsers.add_parser(
            "construct", help="Construct documentation with a chosen backend")

        self.parent_parser.add_argument(
            "-t", "--translator",
            default="rst",
            help=f"Which translator to use. Can be one of {self.TRANSLATORS} or a path to a script."
        )
        self.parent_parser.add_argument(
            "-f", "--format",
            default="rst",
            help=f"Which file format suffix to use on generated documentation."
        )
        self.parent_parser.add_argument(
            "-O", "--options-file",
            help=f"Path to options JSON with data to use in documentation."
        )
        self.parent_parser.add_argument(
            "input_dir", help="Input directory with XML documentation files."
        )
        self.parent_parser.add_argument(
            "output_dir", help="Output directory to save generated documentation."
        )

        processors.append(self.process)

        self.parser.set_defaults(execute=self.execute)

        self.subparsers = self.parser.add_subparsers(
            title="constructor", dest="constructor", description="The constructor to use.")

    def execute(self, args: Namespace):
        self.parser.print_help()

    def process(self, args: Namespace):
        if not hasattr(args, "input_dir") or not hasattr(args, "output_dir"):
            return args

        docs = xml_parser.parse(args.input_dir)

        options: dict[str, str] = {}

        if args.options_file != None:
            options = util.options.load(args.options_file)

        ctx = context_creator.create(docs, options)

        interpreter = BBCodeInterpreter()

        translator = get_translator(args.translator)

        ctx = context_creator.translate(ctx, interpreter, translator)

        args.ctx = ctx

        return args



from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from typing import Optional, TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from argparse import _SubParsersAction  # type: ignore

type Processor = Callable[[Namespace], Namespace]


class CLICommand(ABC):
    """
    Base class that represents a CLI command or subcommand.

    The `register` method should be implemented to connect
    the logic of this command as an `argparse` parser or
    subparser.
    """

    @abstractmethod
    def register(
        self,
        superparsers: "Optional[_SubParsersAction[ArgumentParser]]" = None,
        parent_parser: Optional[ArgumentParser] = None,
        processors: Optional[list[Processor]] = None
    ):
        """
        Abstract method that takes a `subparsers` instance
        (from `argparse.ArgumentParser.add_subparsers`) and
        should append a subparser to them.

        For a root command implementation, the `subparsers`
        parameter may be null, since it isn't a subparser
        as opposed to a main parser.
        """
        pass

    @abstractmethod
    def execute(self, args: Namespace):
        pass

    def process(self, args: Namespace) -> Namespace:
        return args



import inspect
from os import PathLike
import importlib.util
import sys
from types import ModuleType


def load(name: str, path: str | PathLike[str]):
    """
    Loads and executes a module from the file system based on its
    path.

    If no module is found a `FileNotFoundError` is raised, else,
    it is registered in the `sys.modules` `dict` with the given
    `name`.

    This function returns the loaded module, which can be stored
    in a global variable to be accessible like a staticly imported
    module.
    """

    spec = importlib.util.spec_from_file_location(name, path)

    if spec is None or spec.loader is None:
        raise FileNotFoundError(
            f"Could not load module '{name}' from path '{path}'")

    module = importlib.util.module_from_spec(spec)

    spec.loader.exec_module(module)

    sys.modules[name] = module

    return module


def get_classes(module: ModuleType):
    """
    Returns a `list` with `tuples` storing the name and value of
    the classes of a given `module`.
    """

    return list(filter(
        lambda member_data: member_data[1].__module__ == module.__name__,
        inspect.getmembers(module, inspect.isclass)
    ))


def get_functions(module: ModuleType):
    """
    Returns a `list` with `tuples` storing the name and value of
    the functions of a given `module`.
    """

    return list(filter(
        lambda function_data: function_data[1].__module__ == module.__name__,
        inspect.getmembers(module, inspect.isfunction)
    ))



from os import PathLike
from .rst_syntax_translator import RSTSyntaxTranslator
from godocs.util import module


def get_translator(name: str | PathLike[str]):
    """
    Returns a `SyntaxTranslator` instance based on the `name` passed.

    Currently, the only supported translator is the `rst`.

    If it's desired to get a custom translator, its path can
    be passed in the `name` parameter.

    Custom translator should have a class called `SyntaxTranslator`
    which should extend the base `SyntaxTranslator` class.

    If no `SyntaxTranslator` class is exposed, a `NotImplementedError`
    is raised.
    """

    match name:
        case "rst": return RSTSyntaxTranslator()
        case _:
            try:
                return module.load("translator", name).SyntaxTranslator()
            except AttributeError:
                raise NotImplementedError(
                    "Translator not found. Custom translators should have a SyntaxTranslator class")



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from godocs.translation import ast


class Interpreter(ABC):
    """
    Abstract class for interpreters that convert text into an
    abstract syntax tree (AST), returning a `TagNode`.

    Subclasses must implement the `interpret` method, which
    takes a string input and returns a `TagNode`
    representing the parsed syntax of the text.
    """

    @abstractmethod
    def interpret(self, text: str) -> "ast.TagNode":
        """
        Parse the input `text` and return a `TagNode`
        representing it in an AST.
        """

        pass



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from godocs.translation import ast


class SyntaxTranslator(ABC):
    """
    Abstract base class for translating `ast` nodes.

    This class defines a interface for translating two types of
    nodes into string representations: the `TextNode` and the `TagNode`.

    Subclasses must implement the `translate_text` and `translate_tag` methods
    for defining how those nodes should be translated.
    """

    @abstractmethod
    def translate_text(self, node: "ast.TextNode") -> str:
        """
        Abstract method to translate a `TextNode` into its string representation.
        """

        pass

    @abstractmethod
    def translate_tag(self, node: "ast.TagNode") -> str:
        """
        Abstract method to translate a `TagNode` into its string representation.
        """

        pass



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from godocs.translation.translator import SyntaxTranslator


class Node(ABC):
    """
    Abstract base class representing a node in an abstract syntax tree used
    for syntax translation (AST).

    Subclasses must implement the `translate` method, which defines how the node should be
    translated using a provided `SyntaxTranslator`.

    The general protocol for syntax representation using the `godocs.translation.ast`
    module is described below:

    - The topmost node should be a `TagNode` with name `root`;

    - Tags that reference documentation members should be called `reference` and
    have the type of member they reference annotated in the `type` param.
    The name of the referenced member should be stored in the `name` param.
    In the case the referenced member is an `operator`, it should also have
    its operator stored in the `symbol` param.
    If the referenced member is just a `param`, the its AST tag may be a `code`.

    TODO: finish protocol documentation.
    """

    @abstractmethod
    def translate(self, translator: "SyntaxTranslator") -> str:
        """
        Abstract method to translate this `Node` using a given `translator`.
        """

        pass



from abc import ABC, abstractmethod
from os import PathLike
from typing import Any


type ConstructorContext = dict[str, Any]


class Constructor(ABC):

    @abstractmethod
    def construct(self, context: ConstructorContext, path: str | PathLike[str]):
        pass



from os import PathLike
from godocs.util import module
from warnings import deprecated


@deprecated("Old way of choosing constructor implementation. Use plugins instead.")
def get_constructor(name: str | PathLike[str]):
    """
    Returns a `Constructor` instance based on the `name` passed.

    Currently, the only supported constructors is the `jinja`.

    If it's desired to get a custom constructor, its path can
    be passed in the `name` parameter.

    Custom constructors should have a class called `Constructor`
    which should extend the base `Constructor` class.

    If no `Constructor` class is exposed, a `NotImplementedError`
    is raised.
    """

    match name:
        case "jinja": return  # JinjaConstructor()
        case _:
            try:
                return module.load("constructor", name).Constructor()
            except AttributeError:
                raise NotImplementedError(
                    "Constructor not found. Custom constructors should have a Constructor class")



from functools import reduce
from textwrap import indent
import re
from typing import TYPE_CHECKING

from godocs.translation.translator.syntax_translator import SyntaxTranslator
# from godocs.constructor.jinja_constructor.rst.filters import make_code_member_ref

if TYPE_CHECKING:
    from godocs.translation import ast


def normalize_code_member(name: str) -> str:
    # Substitute Array notation from "type[]" to "Array[type]"
    result = re.sub(
        r"(\S+)\[\]",
        lambda match: f"Array[{match.group(1)}]",
        name,
    )
    # Substitute dot notation from "A.B" to "A_B" so it works better
    # in refs and labels.
    result = result.replace('.', '_')

    return result


def make_code_member_label_target(name: str, prefix: str = '') -> str:
    result = f"{prefix + '_' if prefix else ''}"

    result += normalize_code_member(name)

    return result


def make_code_member_ref(full_name: str, prefix: str = '', name: str | None = None) -> str:
    if name == None:
        name = full_name

    return f":ref:`{name} <{make_code_member_label_target(full_name, prefix)}>`"


class RSTSyntaxTranslator(SyntaxTranslator):

    def make_directive(
        self,
        name: str,
        args: list[str] | None = None,
        options: dict[str, str] | None = None,
        content: str = '',
    ) -> str:
        if args is None:
            args = []
        if options is None:
            options = {}

        name_output = name

        args_output = reduce(
            lambda prev, next: prev + next + ' ',
            args,
            '',
        )

        options_output = ''

        for i, option in enumerate(options):
            value = options[option]

            options_output += f":{option}: {value}"

            if i < len(options) - 1:
                options_output += "\n"

        if options_output:
            options_output = indent(options_output, "   ")

        content_output = indent(content, "   ")

        result = f".. {name_output}::"

        if args_output:
            result += f" {args_output}"
        if options_output:
            result += f"\n{options_output}"
        if content_output:
            result += f"\n\n{content_output}"

        return result

    def make_codeblock(self, content: str, language: str = '') -> str:
        return self.make_directive("codeblock", [language], {}, content)

    def translate_text(self, node: "ast.TextNode") -> str:
        return node.content

    def translate_tag(self, node: "ast.TagNode") -> str:
        # First of all, translates the children of the node received.
        content = reduce(
            lambda prev, next: prev + next.translate(self),
            node.children,
            '',
        )

        # Depending on the node name, the resultant syntax will change.
        match node.name:
            case "root": return content
            case "bold": return f"**{content}**"
            case "newline": return f"\n"
            case "italic": return f"*{content}*"
            case "paragraph": return content
            case "code": return f"``{content}``"
            case "codeblock": return self.make_codeblock(
                content,
                node.params.get("language", '')
            )
            case "link": return f"{content} <{node.params.get("url", '')}>_"
            case "reference": return make_code_member_ref(node.params.get("name", ''))
            case _: return ''

        return content



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from godocs.cli import AppCommand


class Plugin(ABC):
    """
    This class allows the extension of the Godocs package
    by other packages that use the
    `[project.entry-points."godocs.plugins"]` field in their
    `pyproject.toml` to expose their `Plugin` implementation.
    """

    @abstractmethod
    def register(self, app: "AppCommand"):
        """
        This abstract method should be implemented to register
        new plugins to the Godocs package.
        """

        pass



from .syntax_translator import SyntaxTranslator
from .rst_syntax_translator import RSTSyntaxTranslator
from .mapper import get_translator

__all__ = ["SyntaxTranslator", "RSTSyntaxTranslator", "get_translator"]



from importlib.metadata import entry_points
from typing import Any


def load() -> list[Any]:
    """
    Loads all plugin entry points registered under the `godocs.plugin`
    group and instantiates them, storing their result in the
    returned `list`.
    """

    plugins: list[Any] = []

    for ep in entry_points(group="godocs.plugins"):
        Plugin = ep.load()

        plugins.append(Plugin())

    return plugins



from textwrap import indent
from typing import TYPE_CHECKING

from .node import Node

if TYPE_CHECKING:
    from godocs.translation.translator import SyntaxTranslator


class TagNode(Node):
    """
    Class representing a node that can hold other `Nodes` as content
    in an abstract syntax tree used for syntax translation (AST).

    Besides its `children`, the `TagNode` class can have a `name` -
    to identify its type - and `params` - to store specific data.
    All of these properties are expected to be passed on construction.
    """

    def __init__(
        self,
        name: str,
        children: list[Node] | None = None,
        params: dict[str, str] | None = None,
    ):
        if children is None:
            children = []
        if params is None:
            params = {}

        self.name = name
        self.children = children
        self.params = params

    def translate(self, translator: "SyntaxTranslator") -> str:
        return translator.translate_tag(self)

    def stringify_params(self) -> str:
        result = ""

        for i, key in enumerate(self.params):
            value = self.params[key]

            result += f"{key}={value}"

            if i < len(self.params) - 1:
                result += ", "

        return result

    def stringify_children(self) -> str:
        result = ""

        for i, child in enumerate(self.children):
            result += str(child)

            if i < len(self.children) - 1:
                result += ",\n"

        return result

    def __str__(self) -> str:
        result = f"<{self.name}"

        if not len(self.params) == 0:
            result += f" {self.stringify_params()}"

        if not len(self.children) == 0:
            result += indent(f"\n{self.stringify_children()}\n", "\t")

        result += '>'

        return result



from argparse import ArgumentParser, Namespace
from typing import TypedDict, cast, Optional, TYPE_CHECKING, Callable
from os import PathLike
from godocs.cli.command.cli_command import CLICommand
from godocs.cli.command.contruct_command import ConstructCommand
from godocs.plugin import Plugin as PluginType, load as load_plugins
from godocs.util import module

if TYPE_CHECKING:
    from argparse import _SubParsersAction  # type: ignore
    from godocs.cli.command.cli_command import Processor


class AppSubcommands(TypedDict):
    construct: ConstructCommand


class AppCommand(CLICommand):
    """
    The main `CLICommand` for the `godocs` app.

    This command exposes as its main (and only, for now) option the
    `"construct"` subcommand, which triggers the generation of
    documentation output.

    It's possible to extend the functionality of this CLI app
    by providing a path to a script in the `--plugin` or `-p`
    option.
    That script should have a function called `register` with
    its signature expecting an `AppCommand` as its parameter.

    That function can then make any modifications/ additions
    to the application parsers and subparsers as needed.
    """

    parser: ArgumentParser
    """
    The `argparse.ArgumentParser` instance this `AppCommand` uses.
    """

    subparsers: "_SubParsersAction[ArgumentParser]"
    """
    The `subparsers` of the `parser` of this `AppCommand`.
    """

    subcommands: AppSubcommands = {
        "construct": ConstructCommand()
    }
    """
    The subcommands this `AppCommand` exposes.

    Currently, there's only the `"construct"` option.
    """

    processors: list[Callable[[Namespace], Namespace]] = []

    def register(
        self,
        superparsers: "Optional[_SubParsersAction[ArgumentParser]]" = None,
        parent_parser: Optional[ArgumentParser] = None,
        processors: "Optional[list[Processor]]" = None
    ):
        """
        Creates the `parser` for this `AppCommand` and
        registers the `--plugin` or `-p` option, as well
        as sets the help printing function as the default
        behavior when nothing else is chosen.
        """

        self.parser = ArgumentParser(
            description="Godot Docs generator CLI")

        self.parser.set_defaults(execute=self.execute)

        self.subparsers = self.parser.add_subparsers(
            title="command", description="The command to execute.")

        self._register_subcommands()

        plugins = load_plugins()

        for p in plugins:
            self._register_plugin(p)

    def execute(self, args: Namespace):
        self.parser.print_help()

    def parse(self):
        args, _ = self.parser.parse_known_args()

        return args

    def start(self, args: Namespace):
        for processor in self.processors:
            args = processor(args)

        args.execute(args)

    def _register_plugin(self, plugin: str | PathLike[str] | PluginType):
        """
        Executes the registering logic for a `plugin` received, giving
        it this `AppCommand` instance so that it can customize it
        as necessary.
        """

        if not isinstance(plugin, PluginType):
            plugin_module = module.load("plugin", plugin)

            Plugin = dict(
                module.get_classes(plugin_module)).get("Plugin")

            if Plugin is None:
                raise NotImplementedError(
                    f"Plugin {plugin} needs to implement a Plugin class")

            plugin = cast(PluginType, Plugin())

        plugin.register(self)

    def _register_subcommands(self):
        """
        Registers the subcommands for this `AppCommand`.
        """

        # Registers the construct command to the subparsers
        self.subcommands["construct"].register(
            self.subparsers, None, self.processors)



from godocs.cli.command import AppCommand


def main():
    """
    Entrypoint for the `godocs` CLI application.
    """

    # Instantiates main app
    app = AppCommand()

    app.register()

    args = app.parse()

    app.start(args)


if __name__ == "__main__":
    main()



from pathlib import Path
import xml.etree.ElementTree as ET

type XMLNode = ET.Element[str]

type XMLDoc = ET.ElementTree[XMLNode]


def parse_file(path: str | Path) -> XMLDoc:
    """
    Parses an XML file from a given path and returns an ElementTree object.

    Args:
      path: Path to the XML file.

    Returns:
      Parsed ElementTree containing the XML data.

    Raises:
        FileNotFoundError: If the file at the given path does not exist.
        xml.etree.ElementTree.ParseError: If the file is not a valid XML document.
    """

    return ET.parse(path)


def parse_folder(path: str | Path) -> list[XMLDoc]:
    """
    Parses all XML files from a given path and returns a list with ElementTree objects.

    Args:
      path: Path to the folder containing XML files.

    Returns:
      List with the parsed ElementTrees containing the XML data.

    Raises:
        NotADirectoryError: If the path doesn't point to a directory.
    """

    path = Path(path)

    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a directory")

    return [parse_file(subpath) for subpath in path.glob("*.xml")]


def parse(path: str | Path) -> list[XMLDoc]:
    """
    Parses one or more XML files from a given path.

    If the path points to a file, parses that file.
    If the path points to a directory, parses all XML files in the directory.
    Returns a list of parsed ElementTree objects.

    Args:
      path: Path to an XML file or a directory containing XML files.

    Returns:
      A list of ElementTree objects parsed from the XML files.

    Raises:
        FileNotFoundError: If the file at the given path does not exist.
        xml.etree.ElementTree.ParseError: If the file is not a valid XML document.
    """

    path = Path(path)

    return [parse_file(path)] if path.is_file() else parse_folder(path)



from os import PathLike
from pathlib import Path
from typing import Callable


def get_subitems(
    path: str | PathLike[str],
    exclude: list[str] | None = None,
    include: list[str] | None = None,
    predicate: Callable[[Path], bool] | None = None,
) -> list[Path]:
    """
    Returns a `list` with the `Paths` of the subitems
    (both files and or directories) of the `path` passed,
    excluding the ones from the `exclude` list and
    including only the ones from the `include` list (if any).
    """

    if exclude is None:
        exclude = []
    if include is None:
        include = []
    if predicate is None:
        predicate = (lambda _: True)

    path = Path(path)

    subitems: list[Path] = []

    for p in path.iterdir():
        if p.name in exclude:
            continue
        if len(include) > 0 and not p.name in include:
            continue
        if not predicate(p):
            continue

        subitems.append(p)

    return subitems


def get_subdirs(
    path: str | PathLike[str],
    exclude: list[str] | None = None,
    include: list[str] | None = None,
    predicate: Callable[[Path], bool] | None = None,
) -> list[Path]:
    """
    Returns a `list` with the `Paths` of the subdirectories of
    the `path` passed, excluding the ones from the `exclude`
    list and including only the ones from the `include` list
    (if any).
    """

    return get_subitems(
        path,
        exclude,
        include,
        predicate=lambda p: p.is_dir() and (
            predicate(p) if predicate is not None else True)
    )



import re
from typing import TypedDict

from .interpreter import Interpreter
from godocs.translation import ast


class TagOptions(TypedDict):
    list: list[str]
    map: dict[str, str]


class BBCodeInterpreter(Interpreter):
    """
    `BBCodeInterpreter` is a specialized class that parses `BBCode`
    markup from Godot docs into an abstract syntax tree (AST) representation.

    To know the list of tags this interpreter supports see the `parse_tag`
    and `parse_element` methods.

    The conversion realized by this class takes into account both standalone
    and nested tags, and supports parsing of tag options and parameters.
    """

    el_regex = r"\[(?P<name>\w+)(?P<options>[\S\s]*?)\](?:(?P<content>[\S\s]*?)\[(?P<closing>\/)\1\])?"
    """
    A regex for capturing `BBCode` element or tags and capturing important information such as
    their `name`, `options`, `content` and `closing` tag.
    """

    def interpret(self, text: str) -> ast.TagNode:
        """
        Parses the input `BBCode` text into an Abstract Syntax Tree
        with a root `TagNode`.
        """

        return self.parse_text(text)

    def parse_options(self, options: str) -> TagOptions:
        """
        Parses a string with the options of a `BBCode` tag and
        returns a `TagOptions` dictionary separating those options
        into a `"list"` (for positional options) and a
        `"map"` (for key-value options).

        Examples:

            parse_options("=https://github.com/nadjiel")
            # Returns:
            # {
            #     "list": [],
            #     "map": {
            #         "": "=https://github.com/nadjiel"
            #     }
            # }

            parse_options(" width=32 height=16")
            # Returns:
            # {
            #     "list": [],
            #     "map": {
            #         "width": "32",
            #         "height": "16"
            #     }
            # }

            parse_options(" Color.operator *")
            # Returns:
            # {
            #     "list": ["Color.operator", "*"],
            #     "map": {}
            # }
        """

        result: TagOptions = {
            "list": [],
            "map": {},
        }

        pairs = options.split(' ')

        for pair in pairs:
            if not pair:
                continue

            key_value = pair.split('=')
            key = key_value[0]

            if len(key_value) == 1:
                result["list"].append(key)
                continue

            value = key_value[1]

            result["map"][key] = value

        return result

    def parse_reference_tag(self, el_match: re.Match[str]) -> ast.TagNode:
        """
        Parses a regex match of a text with a `BBCode`
        tag referencing some documentation member and returns
        its equivalent in an AST `Node`.

        This method interprets the following tags:
        - `[annotation]`
        - `[constant]`
        - `[enum]`
        - `[member]`
        - `[method]`
        - `[constructor]`
        - `[operator]`
        - `[signal]`
        - `[theme_item]`
        - `[param]`
        - `[<Class>]`

        The returned `TagNode` follows the protocol described in the
        `ast.Node` documentation.
        """

        name = el_match.group("name")
        options = el_match.group("options")

        params = self.parse_options(options)

        result = ast.TagNode("tag")

        # Converts every possible reference to a "reference" Node.
        match name:
            case ("annotation"
                  | "constant"
                  | "enum"
                  | "member"
                  | "method"
                  | "constructor"
                  | "operator"
                  | "signal"
                  | "theme_item"):
                # Register data from references with format [<type> <name>].
                result.name = "reference"
                result.params = {
                    "type": name,
                    "name": params["list"][0],
                }

                # If the reference is an operator ([operator <name> <symbol>]),
                # also register its symbol.
                if name == "operator":
                    result.params["symbol"] = params["list"][1]

                return result
            case _: pass

        # If the reference is a [param <name>], converts to inline code.
        if name == "param":
            return ast.TagNode(
                "code", [ast.TextNode(params["list"][0])]
            )

        # If the reference is none of the above, assumes it is a Class reference.
        return ast.TagNode("reference", [], {
            "type": "class", "name": name
        })

    def parse_tag(self, el_match: re.Match[str]) -> ast.Node:
        """
        Parses a regex match of a text with a `BBCode`
        tag returning its equivalent in an AST `Node`.

        This method is only capable of parsing standalone tags,
        not full elements.

        With that said, the only such tags
        are the `[br]`, the `[lb]` and the `[rb]`,
        which are converted to the more suitable `Node`
        dependeing on the AST protocol.

        For full elements, see `parse_element`.
        """

        name = el_match.group("name")

        match name:
            # Returns a new line tag Node.
            case "br": return ast.TagNode("newline")
            # Returns a text Node with "[".
            case "lb": return ast.TextNode("[")
            # Returns a text Node with "]".
            case "rb": return ast.TextNode("]")
            case _: pass

        # If the tag isn't any of the above, it's assumed it is a reference tag,
        # (tags that point to a Class, a property etc).
        return self.parse_reference_tag(el_match)

    def parse_element(self, el_match: re.Match[str]) -> ast.Node:
        """
        Parses a regex match of a text with a `BBCode` element into
        an Abstract Syntax Tree `Node`.

        This method currently understands the following elements:
        - `[b]`
        - `[i]`
        - `[u]`
        - `[s]`
        - `[color]`
        - `[font]`
        - `[img]`
        - `[url]`
        - `[center]`
        - `[kbd]`
        - `[code]`
        - `[codeblock]`

        Not counting these, this method also interprets the standalone tags
        supported by the `parse_tag` method, if the passed `el_match`
        doesn't represent a full element.
        """

        # If the match doesn't have a closing tag, it is a standalone tag.
        if not el_match.group("closing"):
            return self.parse_tag(el_match)

        name = el_match.group("name")
        content = el_match.group("content")
        options = el_match.group("options")

        params = self.parse_options(options)

        element = ast.TagNode(name)

        match name:
            # Converts general markup to standard tag Nodes.
            # case 'p': element.name = "paragraph"
            case 'b': element.name = "bold"
            case 'i': element.name = "italic"
            case 'u': element.name = "underline"
            case 's': element.name = "strikethrough"
            case "color":
                element.name = "color"
                element.params["value"] = params["map"].get('', '')
            case "font":
                element.name = "font"
                element.params["url"] = params["map"].get('', '')
            case "img":
                element.name = "image"
                element.params["width"] = params["map"].get("width", '')
            case "url":
                element.name = "link"
                element.params["url"] = params["map"].get('', content)
            case "center":
                element.name = "alignment"
                element.params["x"] = "center"
            # For the below cases, don't parse the contents, as they aren't meant
            # to be parsed (they are either keyboard keys or code samples).
            case "kbd":
                element.name = "keyboard"
                element.children.append(ast.TextNode(content))
                return element
            case "code":
                element.name = "code"
                element.children.append(ast.TextNode(content))
                return element
            case "codeblock":
                element.name = "codeblock"
                element.params["language"] = params["map"].get("lang", '')
                element.children.append(ast.TextNode(content))
                return element
            case _: pass

        # If there is any content inside the element parsed, use parse_text to
        # parse it and append it to the element as root.
        # Since parse_text may use this own method, this can end up
        # being recursive.
        if content:
            element = self.parse_text(content, element)

        return element

    def parse_text(
        self,
        text: str,
        root: ast.TagNode | None = None,
    ) -> ast.TagNode:
        """
        Parses a `BBCode` text into an Abstract Syntax Tree, with
        its topmost `Node` being a wrapper `"root"` element.

        To know more about the supported nodes, see the `parse_element`
        and `parse_tag` methods.

        To understand how is the structure of the generated AST, see
        the protocol in the `ast.Node` documentation.
        """

        if root is None:
            root = ast.TagNode("root")

        text_start = 0
        text_end = len(text)

        # Searches all elements with the el_regex.
        el_matches = list(re.finditer(
            self.el_regex,
            text,
        ))

        # If no elements are found, the entire content is considered plain text.
        if not el_matches:
            root.children.append(ast.TextNode(text))
            return root

        prev_el_end = text_start

        # For each element found
        for i, el_match in enumerate(el_matches):
            el_start = el_match.start()
            el_end = el_match.end()

            # Parse the content prior to this element, if any.
            if el_start > prev_el_end:
                root.children.append(ast.TextNode(text[prev_el_end:el_start]))

            # Parse the element found, and, consequentially, its children.
            root.children.append(self.parse_element(el_match))

            # Parse the remaining contents, if any.
            if i == len(el_matches) - 1:
                if el_end < text_end:
                    root.children.append(ast.TextNode(text[el_end:]))

            prev_el_end = el_end

        return root



from .node import Node
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from godocs.translation.translator import SyntaxTranslator


class TextNode(Node):
    """
    Class representing a node with text content in an abstract syntax tree used
    for syntax translation (AST).

    This class expects that its text `content` is passed in order
    for it to be constructed.
    """

    def __init__(self, content: str):
        self.content = content

    def translate(self, translator: "SyntaxTranslator") -> str:
        """
        Translates this `TextNode` into a `string` using the provided `SyntaxTranslator`.
        """

        return translator.translate_text(self)

    def __str__(self) -> str:
        return f'"{self.content}"'
